import logging
import traceback
import os
import io
import datetime
import threading
import collections

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

# custom deque implementation to have log line limits as a sliding window that behaves like a stream being consumed when read.
class DequeHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deque = collections.deque(maxlen=100)

    def emit(self, record):
        log_entry = self.format(record)
        self.deque.append(log_entry)

    def read_and_clear(self):
        logs = list(self.deque)
        self.deque.clear()
        return logs

# custom logger implementation that allows for streaming to stdout and tailing of logs per a deque sliding window.
class Logger:
    instance = None
    streamHandler = None
    clientHandler = None
    initialized = False
    lock = threading.Lock()

    def __new__(self, *args, **kwargs):
        with Logger.lock:
            if Logger.instance is None:
                Logger.instance = super().__new__(self, *args, **kwargs)
        return Logger.instance

    def __init__(self, *args, **kwargs):
        with self.lock:
            if not self.initialized:
                self.instance = logging.getLogger("crumbs")

                date_format = "%Y-%m-%d %H:%M:%S %z"
                message_format = "[%(asctime)s] [%(levelname)-5s] [%(filename)13s:%(lineno)-3s] %(message)s"

                formatter = logging.Formatter(fmt=message_format, datefmt=date_format)

                # Check for existing StreamHandler to avoid duplicates
                if not any(isinstance(h, logging.StreamHandler) for h in self.instance.handlers):
                    self.streamHandler = logging.StreamHandler()
                    self.streamHandler.setFormatter(formatter)
                    self.instance.addHandler(self.streamHandler)

                # Check for existing DequeHandler to avoid duplicates
                if not any(isinstance(h, DequeHandler) for h in self.instance.handlers):
                    self.clientHandler = DequeHandler()
                    self.clientHandler.setFormatter(formatter)
                    self.instance.addHandler(self.clientHandler)

                self.instance.propagate = False

                self.initialized = True

    def setLevel(self, level):
        logging.getLogger("crumbs").setLevel(level)
        return

    def info(self, msg, *args, **kwargs):
        self.instance.info(msg, *args, stacklevel=2, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.instance.debug(msg, *args, stacklevel=2, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.instance.warning(msg, *args, stacklevel=2, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.instance.error(msg, *args, stacklevel=2, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.instance.critical(msg, *args, stacklevel=2, **kwargs)

    def tail(self):
        log_entries = self.clientHandler.read_and_clear()
        log_text = '\n'.join(log_entries)

        if log_text:
            log_text += '\n'  # Adds newline character at the end if there is some log_text

        return log_text.encode('utf-8')



import flask
import config
import json
import logger
import sys
import io
import traceback
import contextlib
import urllib3
import time
from huckle import cli, stdin
from flask import Flask, render_template, send_file, jsonify, Response, redirect, url_for, request, render_template_string
import subprocess
import ast

logging = logger.Logger()
logging.setLevel(logger.INFO)

app = None


def get_chat_list():
    try:
        chunks = cli("hai ls --json")
        json_string = ""
        for dest, chunk in chunks:
            if dest == 'stdout':
                json_string += chunk.decode()
        chats = json.loads(json_string)
        # Sort the list with most recent dates first
        sorted_chats = sorted(chats, key=lambda x: x['update_time'], reverse=True)
        return sorted_chats
    except Exception as error:
        logging.error(f"Error getting chat list: {error}")
        return []

def parse_context(context_str):
    try:
        context_data = json.loads(context_str)
        return {
            'messages': context_data.get('messages', []),
            'name': context_data.get('name', ''),
            'title': context_data.get('title', '')
        }
    except json.JSONDecodeError as e:
        logging.error(f"Error parsing context JSON: {e}")
        return {'messages': [], 'name': '', 'title': ''}

def webapp():
    app = Flask(__name__)

    @app.route('/')
    def index():
        try:
            # Get the current context
            chunks = cli("hai context --json")
            context_str = ""
            for dest, chunk in chunks:
                if dest == 'stdout':
                    context_str += chunk.decode()

            # Get chat list for sidebar
            chats = get_chat_list()

            # Get model in use
            chunks = cli(f"hai model --json")
            model = ""
            for dest, chunk in chunks:  # Now unpacking tuple of (dest, chunk)
                if dest == 'stdout':
                    model += chunk.decode()

            # Convert to a python list
            model = ast.literal_eval(model)[0]

            # Get models list
            chunks = cli(f"hai model ls --json")
            models = ""
            for dest, chunk in chunks:  # Now unpacking tuple of (dest, chunk)
                if dest == 'stdout':
                    models += chunk.decode()

            # Convert to a python list
            models = ast.literal_eval(models)

            # Parse the context into structured data
            context_data = parse_context(context_str)

            popup_message = request.args.get('popup')  # Get from query param
            return render_template('index.html',
                                    messages=context_data['messages'],
                                    name=context_data['name'],
                                    title=context_data['title'],
                                    chats=chats,
                                    model=model,
                                    models=models,
                                    popup_message=popup_message)
        except Exception as error:
            logging.error(traceback.format_exc())
            return render_template('index.html',
                                    messages=[],
                                    name='',
                                    title='',
                                    chats=[],
                                    model=None,
                                    models=[],
                                    popup_message=None)

    @app.route('/chat_history')
    def chat_history():
        try:
            # Get chat list for sidebar
            chats = get_chat_list()

            # Get model in use
            chunks = cli(f"hai model --json")
            model = ""
            for dest, chunk in chunks:  # Now unpacking tuple of (dest, chunk)
                if dest == 'stdout':
                    model += chunk.decode()

            # Convert to a python list
            model = ast.literal_eval(model)[0]

            # Get models list
            chunks = cli(f"hai model ls --json")
            models = ""
            for dest, chunk in chunks:  # Now unpacking tuple of (dest, chunk)
                if dest == 'stdout':
                    models += chunk.decode()

            # Convert to a python list
            models = ast.literal_eval(models)

            popup_message = request.args.get('popup')  # Get from query param
            return render_template('chat_history.html', chats=chats,
                                    model=model,
                                    models=models,
                                    popup_message=popup_message)
        except Exception as error:
            logging.error(traceback.format_exc())

        return render_template('chat_history.html', chats=[],
                                model=None,
                                models=[],
                                popup_message=None)

    # We select and set a chat context
    @app.route('/context/<context_id>')
    def navigate_context(context_id):
        try:
            logging.info("Switching to context_id " + context_id)

            chunks = cli(f"hai set {context_id}")
            stderr_output = ""
            stdout_output = ""
            for dest, chunk in chunks:
                if dest == 'stderr':
                    stderr_output += chunk.decode()
                elif dest == 'stdout':
                    stdout_output += chunk.decode()

            if stderr_output:
                logging.error(f"{stderr_output}")
                return redirect(url_for('index', popup=f"{stderr_output}"))

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Context switch failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We delete a chat context with hai rm
    @app.route('/context/<context_id>', methods=['POST'])
    def delete_context(context_id):
        try:
            logging.info("Removing context_id " + context_id)

            chunks = cli(f"hai rm {context_id}")
            for dest, chunk in chunks:
                pass

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Context deletion failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We stream chat data to hai
    @app.route('/chat', methods=['POST'])
    def chat():
        try:
            message = request.form.get('message')
            logging.info("Sending message to selected model...")
            stream = io.BytesIO(message.encode('utf-8'))

            with stdin(stream):
                chunks = cli(f"hai")
                stderr_output = ""
                stdout_output = ""
                for dest, chunk in chunks:
                    if dest == 'stderr':
                        stderr_output += chunk.decode()
                    elif dest == 'stdout':
                        stdout_output += chunk.decode()

                if stderr_output:
                    logging.error(f"{stderr_output}")
                    return redirect(url_for('index', popup=f"{stderr_output}"))

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Context switch failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We set the model with hai model set
    @app.route('/set_model', methods=['POST'])
    def set_model():
        try:
            model = request.form.get('model')
            logging.info(f"Setting model to {model}")

            chunks = cli(f"hai model set {model}")
            for dest, chunk in chunks:
                pass

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Model switch failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We create a new hai chat context with hai new
    @app.route('/new_chat', methods=['POST'])
    def new_chat():
        try:
            chunks = cli("hai new")
            stderr_output = ""
            stdout_output = ""
            for dest, chunk in chunks:
                if dest == 'stderr':
                    stderr_output += chunk.decode()
                elif dest == 'stdout':
                    stdout_output += chunk.decode()

            if stderr_output:
                logging.error(f"{stderr_output}")
                return redirect(url_for('index', popup=f"{stderr_output}"))

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"New chat creation failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We start vibing
    @app.route('/vibestart', methods=['POST'])
    def vibe_start():
        try:
            chunks = cli("hai vibe start")
            for dest, chunk in chunks:
                pass

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Vibing failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    # We stop vibing
    @app.route('/vibestop', methods=['POST'])
    def vibe_stop():
        try:
            chunks = cli("hai vibe stop")
            for dest, chunk in chunks:
                pass

            return redirect(url_for('index'))
        except Exception as error:
            logging.error(f"Vibing failed: {error}")
            logging.error(traceback.format_exc())
            return redirect(url_for('index'))

    @app.route('/vibe_status', methods=['GET'])
    def vibe_status():
        chunks = cli("hai vibe status")
        status_str = ""
        for dest, chunk in chunks:
            if dest == 'stdout':
                status_str += chunk.decode()
        return status_str

    @app.route('/manifest.json')
    def serve_manifest():
        return app.send_static_file('manifest.json')

    return app

app = webapp()



from subprocess import call

import sys

from . import package
from . import config
from . import hutils


def main():
    if len(sys.argv) == 2:

        if sys.argv[1] == "--version":
            show_dependencies()

        elif sys.argv[1] == "help":
            display_man_page(config.manpage_path)
            sys.exit(0)

        elif sys.argv[1] == "path":
            print(config.root)
            sys.exit(0)

        elif sys.argv[1] == "run":
            print('gunicorn --preload --workers=1 --threads=100 -b 127.0.0.1:8000 --chdir `haillo path` "haillo:webapp()"')
#             print('waitress-serve --listen=127.0.0.1:8000 --threads=100 haillo:app')
#             print('gunicorn --workers=1 --threads=100 --timeout=30 --keep-alive=5 --graceful-timeout=10 -b 127.0.0.1:8000 "haillo:webapp()"')
#             print('waitress-serve --max-request-header-size=262144 --clear-untrusted-proxy-headers --listen=127.0.0.1:8000 --channel-timeout=30 --asyncore-use-poll --cleanup-interval=5 haillo:app')
            sys.exit(0)

        else:
            help()

    else:
        help()

def show_dependencies():
    def parse_dependency(dep_string):
        # Common version specifiers
        specifiers = ['==', '>=', '<=', '~=', '>', '<', '!=']

        # Find the first matching specifier
        for specifier in specifiers:
            if specifier in dep_string:
                name, version = dep_string.split(specifier, 1)
                return name.strip(), specifier, version.strip()

        # If no specifier found, return just the name
        return dep_string.strip(), '', ''

    dependencies = ""
    for dep in package.dependencies:
        name, specifier, version = parse_dependency(dep)
        if version:  # Only add separator if there's a version
            dependencies += f" {name}/{version}"
        else:
            dependencies += f" {name}"

    print(f"haillo/{package.__version__}{dependencies}")

def help():
    hutils.eprint("for help, use:\n")
    hutils.eprint("  haillo help")
    sys.exit(2)

# displays a man page (file) located on a given path
def display_man_page(path):
    call(["man", path])



from __future__ import absolute_import, division, print_function

__version__ = "0.4.0"
dependencies = ["flask==3.1.0",
                "huckle>=5.7.0,<6.0.0",
                "hcli_hai>=3.4.2,<4.0.0"]
